unit TedMLPNet;

// Neural MLP network
//
// (c)2003, TEDSoft GmbH, Bonn, Germany
//
// www.tedsoft.de, info@tedsoft.de
//
// Alle Rechte vorbehalten.
// 
// All rights reserved - unless granted by written agreement.
//
// It is not permitted to remove this notice.
//
// Free for noncommercial use.

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Grids, TedNetUnit;

type
  //set of possible transfunctions in the specific layer (none in input layer)
  TTransferFunction = (tanhy, fermi, id);

  //neuron data structure for all layers except for input layer
  //each neuron has weights for each neuron in the preceding layer
  TNeuron = record
              weights: TDoubleArray;
            end;

  TLayer = array of TNeuron;

  

  TTedMLPNet = class(TTedNet)
  private
    FMomentum: double;    //momentum for learning

    //learning etas for each layer
    FEtaHiddenLayer: double;
    FEtaHiddenLayer2: double;
    FEtaOutputLayer: double;

    FWeightMin,FWeightMax: double;  //boundaries for weights
    FSeedWeights: Integer;  //for random number generator
    FTopology: TExtendedTopology;

    FOldError: double; //error of preceding
                       //learning cycle to check for diverging
    FDivergeCounter: integer; //counts length of a series of increasing errors

    //data structures holding the nets weights
    FHiddenLayer, FHiddenLayer2, FOutputLayer: TLayer;

    //data structure for learning
    FDiffWeight: array of array of array of double;

    //function types used in the specific layers
    FTransFuncHiddLayer: TTransferFunction;
    FTransFuncHiddLayer2: TTransferFunction;
    FTransFuncOutputLayer: TTransferFunction;

    //improve weights with FTrainingPattern[indexIn]
    procedure learnWithSingleTrial(const indexIn: integer);

    //returns transferFunc(x)
    function transferOperat(x: double;transferFunc: TTransferFunction): double;

    //output of all layers
    procedure getCompleteOutput(const input: TDoubleArray; var
              OutputHidden,OutputHidden2,OutputNet: TDoubleArray);

    function TanHyperb(x: double): double;
    function FFermi(x: double): double;

    //procedures for changing the properties
    procedure SetMomentum(const momentumIn: double);
    procedure SetTransFuncHiddLayer(const transFuncIn: TTransferFunction);
    procedure SetTransFuncHiddLayer2(const transFuncIn: TTransferFunction);
    procedure SetTransFuncOutputLayer(const transFuncIn: TTransferFunction);
    procedure SetEtaHiddenLayer(const learningRateIn: double);
    procedure SetEtaHiddenLayer2(const learningRateIn: double);
    procedure SetEtaOutputLayer(const learningRateIn: double);
    procedure SetWeightMax(const WeightMaxIn: double);
    procedure SetWeightMin(const WeightMinIn: double);
    procedure SetSeedWeights(const SeedWeightsIn: Integer);
    function GetTrainingPattern(): TTrainingPattern;
    procedure SetTrainingPattern(const trainingPatternIn: TTrainingPattern);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy(); override;

    //returns associated topology
    function getTopologyAsTTopology: TTopology; override;

    function isInit(): boolean;
    procedure SetTrainingPatternWithDoubles(const doubles: TDoubleArray; var
                        numberOfTrainingValues: integer;
                        var trainingInit: boolean);

    //initialises the net
    procedure init();

    //delivers error of net concerning trainingpattern
    function errorOfTrainingData(): double;

    //learning until abortBound (if set) is reached
    //numberOfCycles gives maximum number of learning rounds and gets actual
    //number of learning rounds
    procedure Learning(var numberOfCycles: Integer; const divergeFlag: boolean;
                       var diverging: boolean;
                        const abortBound: double=-1;const seed: integer=-1);

    //delivers net output
    procedure getOutput(const input: TDoubleArray; var OutputNet:
                                TDoubleArray);

  published
    property Momentum: double read FMomentum write SetMomentum;
    property Topology: TExtendedTopology read FTopology;
    property WeightMin: double read FWeightMin write SetWeightMin;
    property WeightMax: double read FWeightMax write SetWeightMax;
    property SeedWeights: Integer read FSeedWeights write SetSeedWeights;
    property TransFuncHiddLayer: TTransferFunction
              read FTransFuncHiddLayer write SetTransFuncHiddLayer;
    property TransFuncHiddLayer2: TTransferFunction
              read FTransFuncHiddLayer2 write SetTransFuncHiddLayer2;
    property TransFuncOutputLayer: TTransferFunction
              read FTransFuncOutputLayer write SetTransFuncOutputLayer;
    property LearningRateHiddenLayer: double read FEtaHiddenLayer
              write SetEtaHiddenLayer;
    property LearningRateHiddenLayer2: double read FEtaHiddenLayer2
              write SetEtaHiddenLayer2;
    property LearningRateOutputLayer: double read FEtaOutputLayer
              write SetEtaOutputLayer;
    property TrainingPattern: TTrainingPattern read GetTrainingPattern
              write SetTrainingPattern;
  end;
//end of TmlpNet declaration

  //catch undefined inputs
  procedure errorHandling(errorIndex: byte);

  procedure Register;


implementation

procedure Register;
begin
  RegisterComponents('Tedsoft', [TTedMLPNet]);
end;



 //implementation TTedMLPNet
constructor TTedMLPNet.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FTopology := TExtendedTopology.Create();
  FTopology.SetTedNet(Self);
  
  //defaults
  FWeightMin:=-1;
  FWeightMax:=1;
  FSeedWeights:=-1;
  FSeedWeights:=-1;
  FEtaHiddenLayer:=0.05;
  FEtaHiddenLayer2:=0.05;
  FEtaOutputLayer:=0.05;
  FMomentum:=0;
  FTransFuncHiddLayer:=fermi;
  FTransFuncOutputLayer:=id;
  FTransFuncHiddLayer2:=id;

  //flags for unspecified training, no initialisation of layers so far
  FTrainingInit:=False;
  FNetInit:=False;
end;

function TTedMLPNet.getTopologyAsTTopology: TTopology;
begin
  result:=FTopology;
end;

procedure TTedMLPNet.SetTransFuncHiddLayer(const transFuncIn:
                                                  TTransferFunction);
begin
  if (FTransFuncHiddLayer<>transFuncIn) then
  begin
    FTransFuncHiddLayer:=transFuncIn;
    FNetInit:=False;
  end;
end;

procedure TTedMLPNet.SetTransFuncHiddLayer2(const transFuncIn:
                                                  TTransferFunction);
begin
  if (FTransFuncHiddLayer2<>transFuncIn) and Topology.HiddenLayer2Active then
  begin
    FTransFuncHiddLayer2:=transFuncIn;
    FNetInit:=False;
  end;
end;

procedure TTedMLPNet.SetTrainingPatternWithDoubles(const doubles: TDoubleArray;
                                          var numberOfTrainingValues: integer;
                                          var trainingInit: boolean);
begin
  FTrainingPattern.SetPatternWithDoubles(doubles,numberOfTrainingValues,
              trainingInit);
end;


procedure TTedMLPNet.SetTransFuncOutputLayer(const transFuncIn:
                                                  TTransferFunction);
begin
  if FTransFuncOutputLayer<>transFuncIn then
  begin
    FTransFuncOutputLayer:=transFuncIn;
    FNetInit:=False;
  end;
end;

procedure TTedMLPNet.SetTrainingPattern
                (const trainingPatternIn: TTrainingPattern);
begin
  FTrainingPattern.assign(trainingPatternIn);
end;

procedure TTedMLPNet.SetEtaHiddenLayer(const learningRateIn: double);
begin
  if (learningRateIn > 0) and (FEtaHiddenLayer <> learningRateIn) then
  begin
    FEtaHiddenLayer:=learningRateIn;
  end;
end;

procedure TTedMLPNet.SetEtaHiddenLayer2(const learningRateIn: double);
begin
  if (learningRateIn > 0) and (FEtaHiddenLayer2 <> learningRateIn)
    and FTopology.HiddenLayer2Active then
  begin
    FEtaHiddenLayer2:=learningRateIn;
  end;
end;

procedure TTedMLPNet.SetEtaOutputLayer(const learningRateIn: double);
begin
  if (learningRateIn > 0) and (FEtaOutputLayer <> learningRateIn) then
  begin
    FEtaOutputLayer:=learningRateIn;
  end;
end;

procedure TTedMLPNet.SetWeightMin(const WeightMinIn: double);
begin
  if ((WeightMinIn < 0) and (WeightMinIn >= -1)) then
  begin
    FWeightMin:=WeightMinIn;
  end;
end;

procedure TTedMLPNet.SetWeightMax(const WeightMaxIn: double);
begin
  if ((WeightMaxIn >= 0) and (WeightMaxIn <= 1)) then
  begin
    FWeightMax:=WeightMaxIn;
  end;
end;

procedure TTedMLPNet.SetSeedWeights(const SeedWeightsIn: Integer);
begin
  FSeedWeights:=SeedWeightsIn;
end;

destructor TTedMLPNet.Destroy();
begin
  FTopology.Free();
  inherited Destroy();
end;

function TTedMLPNet.GetTrainingPattern(): TTrainingPattern;
begin
//  result := TTrainingPattern.create;
//  result.assign(FTrainingPattern);
  result := ftrainingpattern;
end;

procedure TTedMLPNet.init();
var
  i,j,numberOfPrecedingNeurons: integer;
begin
  //catch undefined inputs
  if (FTopology.numberOfInputNeurons < 1)
    or (FTopology.numberOfHiddenNeurons < 1)
    or ((FTopology.numberOfHidden2Neurons < 1) and FTopology.HiddenLayer2Active)
    or (FTopology.numberOfOutputNeurons < 1) then
  begin
    errorHandling(5);
    exit;
  end;

  //get memory according to chosen topology, add biases, initialise the weights

  //hiddenLayer
  SetLength(FHiddenLayer,FTopology.numberOfHiddenNeurons);
  //set weights plus bias
  randSeed:=FSeedWeights;
  for i:=0 to FTopology.numberOfHiddenNeurons-1 do
  begin
    SetLength(FHiddenLayer[i].weights,
        FTopology.numberOfInputNeurons+1);//+1:bias
    for j:=0 to FTopology.numberOfInputNeurons do
    begin
      FHiddenLayer[i].weights[j]:=Random*(weightMax-weightMin)+weightMin;
    end;
  end;

  //second hidden layer (optional)
  if FTopology.HiddenLayer2Active then
  begin
    SetLength(FHiddenLayer2,FTopology.numberOfHidden2Neurons);
    for i:=0 to FTopology.numberOfHidden2Neurons-1 do
    begin
      SetLength(FHiddenLayer2[i].weights,FTopology.numberOfHiddenNeurons+1);
      for j:=0 to FTopology.NumberOfHiddenNeurons do
      begin
        FHiddenLayer2[i].weights[j]:=Random*(weightMax-weightMin)+weightMin;
      end;
    end;

    //for initialisation of output layer
    numberOfPrecedingNeurons:=FTopology.numberOfHidden2Neurons;
  end
  else
    numberOfPrecedingNeurons:=FTopology.numberOfHiddenNeurons;


  SetLength(FOutputLayer,FTopology.numberOfOutputNeurons);
  for i:=0 to FTopology.numberOfOutputNeurons-1 do
  begin
    SetLength(FOutputLayer[i].weights,numberOfPrecedingNeurons+1);
    for j:=0 to numberOfPrecedingNeurons do
    begin
      FOutputLayer[i].weights[j]:=Random*(weightMax-weightMin)+weightMin;
    end;
  end;

  //for learning routine
  if FTopology.HiddenLayer2Active then
  begin
    setLength(FDiffWeight,3);
    setLength(FDiffWeight[0],FTopology.NumberOfHiddenNeurons);
    setLength(FDiffWeight[1],FTopology.NumberOfHidden2Neurons);
    setLength(FDiffWeight[2],FTopology.NumberOfOutputNeurons);
    for i:=0 to FTopology.NumberOfHiddenNeurons-1 do
    begin
      setLength(FDiffWeight[0][i],FTopology.NumberOfInputNeurons+1);//+1:bias
      for j:=0 to FTopology.numberOfInputNeurons do
        FDiffWeight[0][i][j]:=0;
    end;
    for i:=0 to FTopology.NumberOfHidden2Neurons-1 do
    begin
      setLength(FDiffWeight[1][i],FTopology.NumberOfHiddenNeurons+1);//+1:bias
      for j:=0 to FTopology.numberOfHiddenNeurons do
        FDiffWeight[1][i][j]:=0;
    end;
    for i:=0 to FTopology.NumberOfOutputNeurons-1 do
    begin
      setLength(FDiffWeight[2][i],FTopology.NumberOfHidden2Neurons+1);//+1:bias
      for j:=0 to FTopology.NumberOfHidden2Neurons do
        FDiffWeight[2][i][j]:=0;
    end;
  end
  else
  begin
    setLength(FDiffWeight,2);
    setLength(FDiffWeight[0],FTopology.NumberOfHiddenNeurons);
    setLength(FDiffWeight[1],FTopology.NumberOfOutputNeurons);
    for i:=0 to FTopology.NumberOfHiddenNeurons-1 do
    begin
      setLength(FDiffWeight[0][i],FTopology.NumberOfInputNeurons+1);//+1:bias
      for j:=0 to FTopology.numberOfInputNeurons do
        FDiffWeight[0][i][j]:=0;
    end;
    for i:=0 to FTopology.NumberOfOutputNeurons-1 do
    begin
      setLength(FDiffWeight[1][i],FTopology.NumberOfHiddenNeurons+1);//+1:bias
      for j:=0 to FTopology.numberOfHiddenNeurons do
        FDiffWeight[1][i][j]:=0;
    end;
  end;
  FOldError:=-1;
  FDivergeCounter:=0;
  FNetInit:=true;
end;

procedure TTedMLPNet.SetMomentum(const momentumIn: double);
begin
  if (momentumIn <> FMomentum) and (momentumIn >= 0) and (momentumIn < 1) then
    FMomentum:=momentumIn;
end;

procedure TTedMLPNet.Learning(var numberOfCycles: Integer;
                        const divergeFlag: boolean; var diverging: boolean;
                        const abortBound: double=-1;const seed: Integer=-1);
var
  indexArray: array of integer;
  i,fortune,index,cycleIndex: integer;
  numberOfTrials: integer;
  error: double;
begin
  //catch undefined inputs
  if not FNetInit then
  begin
    errorHandling(1);
    numberOfCycles:=0;
    exit;
  end;
  if numberOfCycles < 0 then
  begin
    errorHandling(2);
    exit;
  end;
  if not FtrainingInit then
  begin
    errorHandling(3);
    exit;
  end;

  error:=errorOfTrainingData();
  cycleIndex:=0;

  while (cycleIndex < numberOfCycles) and (error > abortBound) do
  begin
    //learning with each training value, independent of the others
    //select training value randomly

    //use index array, index is chosen randomly, swapped with the last index
    //then index array is reduced by one
    numberOfTrials:=FTrainingPattern.getNumberOfTrainingValues();
    SetLength(indexArray,numberOfTrials);
    for i:=0 to numberOfTrials-1 do
    begin
      indexArray[i]:=i;
    end;
    randSeed:=seed;
    for i:=0 to numberOfTrials-1 do
    begin
      fortune:=Random(Length(indexArray));
      index:=indexArray[fortune];
      learnWithSingleTrial(index);
      indexArray[fortune]:=indexArray[Length(indexArray)-1];
      SetLength(indexArray,Length(indexArray)-1);
    end;
    cycleIndex:=cycleIndex+1;
    error:=errorOfTrainingData;

    //check for diverging if not first cycle overall
    if divergeFlag and (FOldError <> -1) then
    begin
      if error > FOldError then
        FDivergeCounter:=FDivergeCounter+1
      else
        FDivergeCounter:=0; //series of increasing errors ended
    end;
    FOldError:=error;
  end;
  numberOfCycles:=cycleIndex;

  //net has diverged
  if (FDivergeCounter >=4) then
    diverging:=true
  else
    diverging:=false;
end;

procedure errorHandling(errorIndex: byte);
begin
  case errorIndex of
    1: showMessage('MLPnet is not initialised');
    2: showMessage('Number of Learning Cycles is negative');
    3: showMessage('No training pattern initialised');
    4: showMessage('Lengths of arrays is not adequate');
    5: showMessage('Check your topology');
    6: showMessage('Invalid training data');
    7: showMessage('No net available');
  end;
end;

procedure TTedMLPNet.getCompleteOutput(const input: TDoubleArray; var
              OutputHidden,OutputHidden2,OutputNet: TDoubleArray);
var
  i,j: integer;
  sum: double;
  OutputLastHidden: TDoubleArray;
begin
  outputLastHidden:=nil;  //will point on first hidden layer or second
                          //hidden layer (if active)
  //catch undefined inputs
  if (length(input)<FTopology.NumberOfInputNeurons) or
     (length(outputHidden)<FTopology.NumberOfHiddenNeurons) or
    (length(outputNet)<FTopology.NumberOfOutputNeurons) or
    ((length(outputHidden2)<FTopology.NumberOfHidden2Neurons) and
     FTopology.HiddenLayer2Active) then
  begin
    errorhandling(4);
    exit;
  end;

  //output of hiddenLayer
  for i:=0 to FTopology.numberOfHiddenNeurons-1 do
  begin
    sum:=0;
    for j:=0 to FTopology.numberOfInputNeurons-1 do
    begin
      sum:=sum + FHiddenLayer[i].weights[j]*input[j];
    end;
    //bias of inputLayer
    sum:=sum + FHiddenLayer[i].weights[FTopology.numberOfInputNeurons];

    OutputHidden[i]:=transferOperat(sum,FTransFuncHiddLayer);
  end;

  //output of hiddenLayer2 if active
  if FTopology.HiddenLayer2Active then
  begin
    for i:=0 to FTopology.numberOfHidden2Neurons-1 do
    begin
      sum:=0;
      for j:=0 to FTopology.numberOfHiddenNeurons-1 do
      begin
        sum:=sum + FHiddenLayer2[i].weights[j]*OutputHidden[j];
      end;
      //bias of hiddenLayer
      sum:=sum + FHiddenLayer2[i].weights[FTopology.numberOfHiddenNeurons];

      OutputHidden2[i]:=transferOperat(sum,FTransFuncHiddLayer2);
    end;

    //to link outputLayer with last hidden layer
    OutputLastHidden:=OutputHidden2;
  end
  else
  begin
    //only one hidden layer
    OutputLastHidden:=OutputHidden;
  end;

  //output of the last layer
  for i:=0 to FTopology.NumberOfOutputNeurons-1 do
  begin
    sum:=0;
    for j:=0 to Length(OutputLastHidden)-1 do
    begin
      sum:=sum + FOutputLayer[i].weights[j]*OutputLastHidden[j];
    end;
    //bias of last hidden layer
    sum:=sum + FOutputLayer[i].weights[Length(OutputLastHidden)];

    OutputNet[i]:=transferOperat(sum,FTransFuncOutputLayer);
  end;
end;

function TTedMLPNet.transferOperat(x: double;transferFunc: TTransferFunction)
                      : double;
begin
  case transferFunc of
    tanhy: transferOperat:=TanHyperb(x);
    fermi: transferOperat:=FFermi(x)
  else
    transferOperat:=x;
  end;
end;

function TTedMLPNet.TanHyperb(x: double): double;
var
  temp: double;
begin
  temp := exp(x);
  TanHyperb := ((temp-1/temp)/(temp+1/temp));
end;

procedure TTedMLPNet.learnWithSingleTrial(const indexIn: integer);
var
  input,outputNet,outputHidden,outputHidden2,
    Delta,FormerDelta,SwapDelta: TDoubleArray;
  i,j,k,max,numberOfSuceedingNeurons: integer;
  sum: double;
  suceedingLayer: TLayer;
begin
  suceedingLayer:=nil;
  input:=nil;
  SwapDelta:=nil;

  //catch undefined inputs
  if not FNetInit then
  begin
    errorHandling(1);
    exit;
  end;
  if not FtrainingInit then
  begin
    errorHandling(3);
    exit;
  end;
  input:=FTrainingPattern[indexIn].input;

  //max gets size of biggest layer in the net, get memory for nets output
  max:=FTopology.NumberOfInputNeurons;
  setLength(outputHidden,FTopology.NumberOfHiddenNeurons);
  if FTopology.NumberOfHiddenNeurons > max then
    max:=FTopology.NumberOfHiddenNeurons;
  if FTopology.HiddenLayer2Active then
  begin
    setLength(outputHidden2,FTopology.NumberOfHidden2Neurons);
    if FTopology.NumberOfHidden2Neurons > max then
      max:=FTopology.NumberOfHidden2Neurons;
  end;
  setLength(outputNet,FTopology.NumberOfOutputNeurons);
  if FTopology.NumberOfOutputNeurons > max then
    max:=FTopology.NumberOfOutputNeurons;

  getCompleteOutput(input,OutputHidden,OutputHidden2,OutputNet);

  //backpropagation for weights' improvement

  //get memory for required data structures during learning
  SetLength(Delta,max);
  SetLength(FormerDelta,max);

  //start with weights in output layer
  case FTransFuncOutputLayer of
    //tangens hyperbolicus
    tanhy:
    begin
      for i:=0 to FTopology.NumberOfOutputNeurons-1 do
      begin
        Delta[i]:=(1-OutputNet[i]*OutputNet[i])*
                  (FTrainingPattern[indexIn].output[i]-OutputNet[i]);

      end;
    end;
    //fermi
    fermi:
    begin
      for i:=0 to FTopology.NumberOfOutputNeurons-1 do
      begin
        Delta[i]:=OutputNet[i]*(1-OutputNet[i])*
                  (FTrainingPattern[indexIn].output[i]-OutputNet[i]);
      end;
    end;
    id:
    begin
      for i:=0 to FTopology.NumberOfOutputNeurons-1 do
      begin
        Delta[i]:=FTrainingPattern[indexIn].output[i]-OutputNet[i];
      end;
    end;
  end;

  //calculate the adjustments of the weights in the output layer
  //using momentum
  for i:=0 to FTopology.NumberOfOutputNeurons-1 do
  begin
    if FTopology.HiddenLayer2Active then
    begin
      for j:=0 to FTopology.numberOfHidden2Neurons-1 do
      begin
        FDiffWeight[2][i][j]:=FEtaOutputLayer*Delta[i]*OutputHidden2[j]+
                                 FMomentum*FDiffWeight[2][i][j];
      end;
      //bias of second hidden layer
      FDiffWeight[2][i][FTopology.numberOfHidden2Neurons]:=
            FEtaOutputLayer*Delta[i]+
            FMomentum*FDiffWeight[2][i][FTopology.numberOfHidden2Neurons];
    end
    else
    begin
      for j:=0 to FTopology.numberOfHiddenNeurons-1 do
      begin
        FDiffWeight[1][i][j]:=FEtaOutputLayer*Delta[i]*OutputHidden[j]+
                                 FMomentum*FDiffWeight[1][i][j];
      end;
      //bias of first hidden layer
      FDiffWeight[1][i][FTopology.numberOfHiddenNeurons]:=
            FEtaOutputLayer*Delta[i]+
            FMomentum*FDiffWeight[1][i][FTopology.numberOfHiddenNeurons];
    end;
  end;
  //learning in the output layer finished

  //calculate Deltas in the hidden layers, first save current Deltas
  SwapDelta:=Delta;
  Delta:=FormerDelta;
  FormerDelta:=SwapDelta;
  if FTopology.HiddenLayer2Active then
  begin
    case FTransFuncHiddLayer2 of
      tanhy:
      begin
        for j:=0 to FTopology.NumberOfHidden2Neurons-1 do
        begin
          sum:=0;
          //calculate Deltas using FormerDeltas and corresponding weights
          //in output layer
          for k:=0 to FTopology.numberOfOutputNeurons-1 do
          begin
            sum:=sum+FormerDelta[k]*FOutputLayer[k].weights[j];
          end;
          Delta[j]:=(1-OutputHidden2[j]*OutputHidden2[j])*sum;

          //adjusting weights
          for k:=0 to FTopology.NumberOfHiddenNeurons-1 do
            FDiffWeight[1][j][k]:=FEtaHiddenLayer2*OutputHidden[k]*
                                    Delta[j]+FMomentum*FDiffWeight[1][j][k];
          //bias of hidden layer
          FDiffWeight[1][j][FTopology.NumberOfHiddenNeurons]:=
                  FEtaHiddenLayer2*Delta[j]+
                  FMomentum*FDiffWeight[1][j][FTopology.NumberOfHiddenNeurons];
        end;
      end;
      fermi:
      begin
        for j:=0 to FTopology.NumberOfHidden2Neurons-1 do
        begin
          sum:=0;
          //calculate Deltas using FormerDeltas and corresponding weights
          //in output layer
          for k:=0 to FTopology.numberOfOutputNeurons-1 do
          begin
            sum:=sum+FormerDelta[k]*FOutputLayer[k].weights[j];
          end;
          Delta[j]:=OutputHidden2[j]*(1-OutputHidden2[j])*sum;

          //adjusting weights
          for k:=0 to FTopology.NumberOfHiddenNeurons-1 do
            FDiffWeight[1][j][k]:=FEtaHiddenLayer2*OutputHidden[k]*
                                    Delta[j]+FMomentum*FDiffWeight[1][j][k];
          //bias of hidden layer
          FDiffWeight[1][j][FTopology.NumberOfHiddenNeurons]:=
                  FEtaHiddenLayer2*Delta[j]+
                  FMomentum*FDiffWeight[1][j][FTopology.NumberOfHiddenNeurons];
        end;
      end;
      id:
      begin
        for j:=0 to FTopology.NumberOfHidden2Neurons-1 do
        begin
          sum:=0;
          //calculate Deltas using FormerDeltas and corresponding weights
          //in output layer
          for k:=0 to FTopology.numberOfOutputNeurons-1 do
          begin
            sum:=sum+FormerDelta[k]*FOutputLayer[k].weights[j];
          end;
          Delta[j]:=sum;

          //adjusting weights
          for k:=0 to FTopology.NumberOfHiddenNeurons-1 do
            FDiffWeight[1][j][k]:=FEtaHiddenLayer2*OutputHidden[k]*
                                    Delta[j]+FMomentum*FDiffWeight[1][j][k];
          //bias of hidden layer
          FDiffWeight[1][j][FTopology.NumberOfHiddenNeurons]:=
                  FEtaHiddenLayer2*Delta[j]+
                  FMomentum*FDiffWeight[1][j][FTopology.NumberOfHiddenNeurons];
        end;
      end;
    end;

    SwapDelta:=Delta;
    Delta:=FormerDelta;
    FormerDelta:=SwapDelta;

    //for link to first hidden layer
    numberOfSuceedingNeurons:=FTopology.NumberOfHidden2Neurons;
    suceedingLayer:=FHiddenLayer2;
  end
  else
  begin
    numberOfSuceedingNeurons:=FTopology.NumberOfOutputNeurons;
    suceedingLayer:=FOutputLayer;
  end;


  //hidden layer
  case FTransFuncHiddLayer of
    tanhy:
    begin
      for j:=0 to FTopology.NumberOfHiddenNeurons-1 do
      begin
        sum:=0;
        for k:=0 to numberOfSuceedingNeurons-1 do
        begin
          sum:=sum+FormerDelta[k]*suceedingLayer[k].weights[j];
        end;
        Delta[j]:=(1-OutputHidden[j]*OutputHidden[j])*sum;

        //adjusting weights
        for k:=0 to FTopology.NumberOfInputNeurons-1 do
          FDiffWeight[0][j][k]:=FEtaHiddenLayer*Input[k]*
                                    Delta[j]+FMomentum*FDiffWeight[0][j][k];
        //bias neuron of input layer
        FDiffWeight[0][j][FTopology.NumberOfInputNeurons]:=FEtaHiddenLayer*
          Delta[j]+FMomentum*FDiffWeight[0][j][FTopology.NumberOfInputNeurons];
      end;
    end;
    fermi:
    begin
      for j:=0 to FTopology.NumberOfHiddenNeurons-1 do
      begin
        sum:=0;
        for k:=0 to numberOfSuceedingNeurons-1 do
        begin
          sum:=sum+FormerDelta[k]*suceedingLayer[k].weights[j];
        end;
        Delta[j]:=OutputHidden[j]*(1-OutputHidden[j])*sum;

        //adjusting weights
        for k:=0 to FTopology.NumberOfInputNeurons-1 do
          FDiffWeight[0][j][k]:=FEtaHiddenLayer*Input[k]*
                                    Delta[j]+FMomentum*FDiffWeight[0][j][k];
        //bias neuron of input layer
        FDiffWeight[0][j][FTopology.NumberOfInputNeurons]:=FEtaHiddenLayer*
          Delta[j]+FMomentum*FDiffWeight[0][j][FTopology.NumberOfInputNeurons];
      end;
    end;
    id:
    begin
      for j:=0 to FTopology.NumberOfHiddenNeurons-1 do
      begin
        sum:=0;
        for k:=0 to numberOfSuceedingNeurons-1 do
        begin
          sum:=sum+FormerDelta[k]*suceedingLayer[k].weights[j];
        end;
        Delta[j]:=sum;

        //adjusting weights
        for k:=0 to FTopology.NumberOfInputNeurons-1 do
          FDiffWeight[0][j][k]:=FEtaHiddenLayer*Input[k]*
                                    Delta[j]+FMomentum*FDiffWeight[0][j][k];
        //bias neuron of input layer
        FDiffWeight[0][j][FTopology.NumberOfInputNeurons]:=FEtaHiddenLayer*
          Delta[j]+FMomentum*FDiffWeight[0][j][FTopology.NumberOfInputNeurons];
      end;
    end;
  end;

  //update weights
  if FTopology.HiddenLayer2Active then
  begin
    for i:=0 to FTopology.NumberOfOutputNeurons-1 do
      for j:=0 to FTopology.NumberOfHidden2Neurons do
      begin
        FOutputLayer[i].weights[j]:=FOutputLayer[i].weights[j]
                                      +FDiffWeight[2][i][j];
      end;
    for i:=0 to FTopology.NumberOfHidden2Neurons-1 do
      for j:=0 to FTopology.NumberOfHiddenNeurons do
      begin
        FHiddenLayer2[i].weights[j]:=FHiddenLayer2[i].weights[j]
                                      +FDiffWeight[1][i][j];
      end;
    for i:=0 to FTopology.NumberOfHiddenNeurons-1 do
      for j:=0 to FTopology.NumberOfInputNeurons do
      begin
        FHiddenLayer[i].weights[j]:=FHiddenLayer[i].weights[j]
                                      +FDiffWeight[0][i][j];
      end;
  end
  else
  begin
    for i:=0 to FTopology.NumberOfOutputNeurons-1 do
      for j:=0 to FTopology.NumberOfHiddenNeurons do
      begin
        FOutputLayer[i].weights[j]:=FOutputLayer[i].weights[j]
                                      +FDiffWeight[1][i][j];
      end;
    for i:=0 to FTopology.NumberOfHiddenNeurons-1 do
      for j:=0 to FTopology.NumberOfInputNeurons do
      begin
        FHiddenLayer[i].weights[j]:=FHiddenLayer[i].weights[j]
                                      +FDiffWeight[0][i][j];
      end;
  end;
end;

function TTedMLPnet.errorOfTrainingData(): double;
var
  error: double;
  trainingIndex,i : integer;
  OutputNet: array of TDoubleArray;
begin
  if (not FNetInit) then
  begin
    errorHandling(1);
    errorOfTrainingData:=0;
    exit;
  end;

  //get datastructure for output
  SetLength(OutputNet,FTrainingPattern.getNumberOfTrainingValues());
  for i:=0 to FTrainingPattern.getNumberOfTrainingValues-1 do
  begin
    SetLength(OutputNet[i], FTopology.NumberOfOutputNeurons);
  end;

  //compare net output with training output
  error:=0;
  for trainingIndex:=0 to FTrainingPattern.getNumberOfTrainingValues()-1 do
  begin
    getOutput(FTrainingPattern[trainingIndex].input,OutputNet[trainingIndex]);
    error:=error+sqrt(DistanceSquare(FTrainingPattern[trainingIndex].output,
                      OutputNet[trainingIndex]));
  end;
  errorOfTrainingData:=error;
end;

procedure TTedMLPnet.getOutput(const input: TDoubleArray; var OutputNet:
                                TDoubleArray);
var
  outputHidden,outputHidden2: TDoubleArray;
begin
  setLength(OutputHidden,FTopology.numberOfHiddenNeurons);
  outputHidden2:=nil;
  if FTopology.HiddenLayer2Active then
    setLength(OutputHidden2,FTopology.numberOfHidden2Neurons);
  getCompleteOutput(input,outputHidden,outputHidden2,outputNet);
end;

function TTedMLPNet.FFermi(x: double): double;
begin
     if (x>1e10) then
       FFermi:=1
     else
       if (x<-1e10) then
         FFermi:=0
       else
         FFermi := 1.0/(1.0+exp(-x));
end;

function TTedMLPnet.isInit(): boolean;
begin
  isInit:=FNetInit;
end;

end.


